require "config"
require "message"
require "signature"
require "util"


class Gateway
  class Error < StandardError
  end

  def initialize
    @datadir = $conf["datadir"]
    @root    = $conf["root"]
    @message = Message.load(ENV["HTTP_ACCEPT_LANGUAGE"])
    @head    = ""
    @body    = ""
  end

  attr_reader :head, :body

  def str_encode(s)
    require "uri"
    return URI.encode(s)
  end

  def str_decode(s)
    require "uri"
    return URI.decode(s)
  end

  def file_encode(type, key)
    enkey = key.unpack("H*")[0].upcase
    return "#{type}_#{enkey}"
  end

  def file_decode(name)
    s    = name.split(/_/, 2)
    name = s[1]
    name = [name].pack("H*")
    return escape(name)
  end

  def rec(hs)
    return Util.rec(hs)
  end

  def xdie(msg)
    @head << "Content-Type: text/plain\r\n" <<
             "X-Shingetsu: #{$conf["version"]}\r\n" <<
             "\r\n"
    @body << msg << "\n"
    raise Gateway::Error
  end

  #
  # escape
  #
  def escape(s)
    s.gsub!(/&/,    "&amp;")
    s.gsub!(/&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);/) {"&#{$1};"}
    s.gsub!(/</,    "&lt;")
    s.gsub!(/>/,    "&gt;")
    s.gsub!(/\r\n/, "<br>")
    s.gsub!(/\r/,   "/"   )
    return s
  end

  #
  # unescape arguments
  #
  def args(str)
    Util.printtrace(str) ## debug

    require 'uri'
    items = str.split(/&/)
    arg = {}
    items.each {|item|
      cd = item.split(/=/)
      if cd.size >= 2 then
        cd[1].gsub!(/\+/, " ")
        arg[cd[0]] = escape(URI.decode(cd[1]))
      else
        arg[cd[0]] = ""
      end
    }
    return arg
  end

  #
  # unescape arguments(multipart/form-data)
  #
  def argsFromMulti(input)
    Util.printtrace(input) ## debug

    arg = {}
    input =~ /\s/
    boundary = $`
    inputs = input.split(boundary)
    inputs.shift

    inputs.each {|item|
      bufs = item.split(/\r\n/)
      next if bufs.size == 1
      x = bufs.shift

      bufs[0] =~ /Content-Disposition: form-data; name="([^"]+)"/i
      key = $1

      if bufs[0] =~ /filename="([^"]+)"/i then
        s = $1.sub(/.*[\/\\]/, "")
        if s =~ /\.([^.]*)$/ then
          s = $1
          s.downcase!
        else
          s = ""
        end
        arg["auto_suffix"] = s
      end

      bufs.shift until bufs.empty? || (bufs[0] == "")
      bufs.shift
      arg[key] = bufs.join("\r\n")
      if key != "attach" then
        arg[key] = escape(arg[key])
      end
    }
    return arg
  end

  #
  # date
  #
  def xlocaltime(sec)
    Util.printtrace(sec) ## debug

    t = Time.at(sec.to_i).localtime
    return t.strftime("%Y/%m/%d %H:%M")
  end

  #
  # make file
  #
  def touch(file)
    Util.printtrace(file) ## debug

    stat = CacheStat.list() # Hash
    unless FileTest.size?("#{@datadir}/#{file}.dat") then
      Util.lock()
      File.open("#{@datadir}/#{file}.dat", "a") {|f|}
      Util.unlock()
      stat = CacheStat.list()
      stat[file] = {}
      stat[file]["records"] = 0
      stat[file]["stamp"  ] = 0
      stat[file]["size"   ] = 0
      CacheStat.sync(stat)
    end
    if FileTest.zero?("#{@datadir}/#{file}.dat") then
      node = Cache.search(file)
      if node then
        Cache.getRegion(file, node)
      end
    end
  end
  
  #
  # read note history
  #
  def readNote(file)
    Util.printtrace(file) ## debug

    buf = ""
    ref = {}
    begin
      IO.foreach("#{@datadir}/#{file}.dat") {|line|
        line.chomp!
        hrec = rec(line)
        next if (hrec.size <= 2) || hrec["remove_stamp"]
        buf = line.dup
        if hrec["base_stamp"] then
          key = hrec["stamp"] + "_" + hrec["id"]
          ref[key] = 0
          key = hrec["base_stamp"] + "_" + hrec["base_id"]
          ref[key] += 1 if ref[key]
        end
      }
    rescue Errno::ENOENT, Errno::EACCES
    end
    orphan = ref.values.grep(0).size
    orphan = 0 if orphan == 1
    return buf, orphan
  end

  #
  # add record
  #
  def addRecord(file, stamp, body)
    Util.printtrace(file, stamp, body) ## debug

    id = Extern.md5digest(body)
    Util.lock()
    begin
      File.open("#{@datadir}/#{file}.dat", "a") {|f|
        f << "#{stamp}<>#{id}<>#{body}\n"
      }
    rescue Errno::ENOENT, Errno::EACCES
      raise "#{@datadir}/#{file}.dat: failed to open. #{$!}"
    end
    Util.unlock()

    stat = CacheStat.list()
    stat[file]["records"] += 1
    stat[file]["stamp"  ] =  stamp
    stat[file]["size"   ] =  FileTest.size?("#{@datadir}/#{file}.dat") || 0
    CacheStat.sync(stat)

    return id
  end

  #
  # post article
  #
  def post(thread, arg)
    Util.printtrace(thread, arg) ## debug

    suffix, attach = "", ""

    arg["name"  ] = "" unless arg["name"  ]
    arg["mail"  ] = "" unless arg["mail"  ]
    arg["passwd"] = "" unless arg["passwd"]

    if arg["attach"] && (arg["attach"] != "") then
      if arg["attach"].size > $conf["filelimit"]*1024*1024 then
        xdie(@message["big_file"])
      end
      attach = Extern.base64encode(arg["attach"])
      if arg["suffix"] != "AUTO" then
        suffix = arg["suffix"]
      elsif arg["auto_suffix"] && (arg["auto_suffix"] != "") then
        suffix = arg["auto_suffix"]
      else
        suffix = "txt"
      end
    end

    stamp = arg["error"] ? Util.time() : Time.now.to_i
    arg["suffix"] = suffix
    arg["attach"] = attach

    body, target = "", ""
    ["base_stamp", "base_id","name", "mail", "body", "suffix", "attach"].each {|s|
      if arg[s] && (arg[s] != "") then
        body   << "<>#{s}:#{arg[s]}"
        target << ",#{s}"
      end
    }

    body.sub!(/^<>/, "")
    target.sub!(/^,/, "")
    xdie(@message["null_article"]) if body == ""
    if arg["passwd"] != "" then
      pubkey, sign = Signature.generate(arg["passwd"], body)
      body = "pubkey:#{pubkey}<>sign:#{sign}<>target:#{target}<>" + body
    end

    id = addRecord(thread, stamp, body)
    if arg["dopost"] then
      Util.addUpdate(stamp, id, thread)
      Cache.tellupdate(thread, stamp, id, nil)
    end
    return id[0, 8] # substr
  end

  #
  # make new list element
  #
  def newElement(file, arg)
    Util.printtrace(file, arg) ## debug

    stamp = Util.time()
    xdie(@message["null_name"]) if arg["name"] == ""
    xdie(@message["bad_name" ]) if arg["name"] =~ /\]\]/
    xdie(@message["null_type"]) if arg["type"] == ""

    body = "type:#{arg["type"]}<>link:#{arg["name"]}"
    id = addRecord(file, stamp, body)
    if arg["dopost"] then
      Util.addUpdate(stamp, id, file);
      Cache.tellupdate(file, stamp, id, nil)
    end
  end
  
  #
  # add new record and marge that and last record (for note)
  #
  def margeRecord(file, arg)
    Util.printtrace(file, arg) ## debug

    stamp      = arg["error"] ? Util.time() : Time.now.to_i
    body       = arg["message"]
    base_stamp = ""
    base_id    = ""
    buf,       = readNote(file)

    if buf != "" then
      hrec = rec(buf);
      Util.debug("hrec=",hrec["body"]) ## debug
      if hrec["body"] != "" then
        if hrec["body"] =~ /<br>$/ then
          body = "#{hrec["body"]}<br>#{body}"
        else
          body = "#{hrec["body"]}<br><br>#{body}"
        end
      end
      base_stamp = hrec["stamp"]
      base_id    = hrec["id"]
    end
    xdie(@message["null_article"]) if body == ""
    body =  "body:#{body}"
    body << "<>base_stamp:#{base_stamp}<>base_id:#{base_id}" if base_stamp

    id = addRecord(file, stamp, body)
    if arg["dopost"] then
      Util.addUpdate(stamp, id, file)
      Cache.tellupdate(file, stamp, id, nil)
    end
    return id[0, 8] # substr
  end
  
  #
  # search string
  # return: filename (Array)
  #
  def search(type, query)
    Util.printtrace(type, query) ## debug

    result = []
    Dir.glob("#{@datadir}/#{type}_*.dat") {|file|
      file2 = file.dup
      file2.sub!(/^#{@datadir}\//, "")
      file2.sub!(/.dat$/, "")
      x = file_decode(file2)
#     y = str_encode(x)
      found = false
      if x =~ /#{query}/ then
        found = true
      else
        IO.foreach(file) {|line|
          if line =~ /#{query}/ then
            found = true
            break
          end
        }
      end
      result << file2 if found
    }
    return result
  rescue Errno::ENOENT, Errno::EACCES
    raise "#{file}: open failed. #{$!}"
  end
  
  #
  # post delete message to other nodes
  #
  def postDeleteMessage(file, stamp, id, arg)
    Util.printtrace(file, stamp, id, arg) ## debug

    arg["name"  ] = "" unless arg["name"  ]
    arg["mail"  ] = "" unless arg["mail"  ]
    arg["passwd"] = "" unless arg["passwd"]
    newstamp = arg["error"] ? Util.time() : Time.now.to_i
    body   = ""
    target = ""
    if arg["name"] && (arg["name"] != "") then
      body   << "<>name:#{arg["name"]}"
      target << ",name"
    end
    # if ((defined $arg{mail})&&($arg{mail} ne "")) {
    #   $body .= "<>mail:$arg{mail}";
    #   $target .= ",mail";
    # }
    if arg["message"] && (arg["message"] != "") then
      body   << "<>body:#{arg["message"]}"
      target << ",body"
    end

    body   << "<>remove_stamp:#{stamp}<>remove_id:#{id}"
    target << ",remove_stamp,remove_id"

p body
    body.sub!(/^<>/, "")
    target.sub!(/^,/, "")

    xdie(@message["null_article"]) if body == ""

    if arg["passwd"] != "" then
      pubkey, sign = Signature.generate(arg["passwd"], body)
      body = "pubkey:#{pubkey}<>sign:#{sign}<>target:#{target}<>" + body
    end
    newid = addRecord(file, newstamp, body)
    Util.addUpdate(newstamp, newid, file)
    Cache.tellupdate(file, newstamp, newid, nil)
  end

  #
  # encode bracket string to link
  #
  def bracket_link(str)
    Util.printtrace(str) ## debug

    link = str.dup
    if link =~ %r!^/(thread)/([^/]+)/([0-9a-f]{8})$! then
      link = "#{@root}#{$1}.cgi/" + str_encode($2) + "#r#{$3}"
      return "<a href='#{link}'>[[#{str}]]</a>"

    elsif link =~ %r!^/(list|thread|note)/([^/]+)$! then
      link = "#{@root}#{$1}.cgi/" + str_encode($2)
      return "<a href='#{link}'>[[#{str}]]</a>"

    elsif link =~ %r!^([^/]+)/([0-9a-f]{8})$! then
      link = str_encode($1) + "#r#{$2}"
      return "<a href='#{link}'>[[#{str}]]</a>"

    elsif link =~ %r!^([^/]+)$! then
      link = str_encode($1)
      return "<a href='#{link}'>[[#{str}]]</a>"

    else
      return "[[#{str}]]"
    end
  end

  #
  # check admin address
  #
  def checkAdmin()
    Util.printtrace() ## debug

    if ENV["REMOTE_ADDR"] !~ /#{$conf["adminAddr"]}/ then
      print403()
      exit
    end
  end

  #
  # print CGI&HTML header
  #
  def printHeader(title)
    Util.printtrace(title) ## debug

    @head << 
        "Content-Type: text/html; charset=UTF-8\r\n"
        "Content-Language: #{@message["lang"]}\r\n" <<
        "X-Shingetsu: #{$conf["version"]}\r\n" <<
        "\r\n"
    @body << 
        '<?xml version="1.0" encoding="UTF-8"?>' <<
        '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"' <<
        ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">' <<
        "<html xmlns='http://www.w3.org/1999/xhtml' xml:lang='#{@message["lang"]}'>" <<
        '<head>' <<
        "<meta http-equiv='Content-Language' content='#{@message["lang"]}' />" <<
        '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />' <<
        "<title>#{title}</title>" <<
        "<link rev='made' href='mailto:fuktommy\@users.sourceforge.net' />" <<
        "<link rel='contents' href='#{@root}list.cgi' />" <<
        "<link rel='stylesheet' type='text/css' href='#{$conf["css"]}' />" <<
        "</head>" <<
        "<body><p class='head'>" <<
        "<a href='#{@root}gateway.cgi'>#{@message["top"]}</a> | " <<
        "<a href='#{@root}list.cgi'>#{@message["menu"]}</a> | " <<
        "<a href='#{@root}gateway.cgi/new'>#{@message["new"]}</a> | " <<
        "<a href='#{@root}gateway.cgi/index'>#{@message["index"]}</a> | " <<
        "<a href='#{@root}gateway.cgi/changes'>#{@message["changes"]}</a> | " <<
        "<a href='#{@root}gateway.cgi/update'>#{@message["update"]}</a> | " <<
        "<a href='#{@root}gateway.cgi/search'>#{@message["search"]}</a>" <<
        "</p>" <<
        "<h1><a href='#{@root}gateway.cgi/search/" << str_encode(title) <<
        "' class='title' title='#{@message["search"]}'>#{title}</a></h1>"
  end

  #
  # print CGI header (403 forbidden)
  #
  def print403()
    Util.printtrace() ## debug

    printHeader(@message["403"])
    @body << "</body></html>\n"
  end

  #
  # print CGI header (404 not found)
  #
  def print404(file = nil)
    Util.printtrace(file) ## debug

    printHeader(@message["404"])
    @body << "<p>#{@message["try_later"]}</p>"
    if file then
      @body << "<form method='get' action='#{ENV["SCRIPT_NAME"]}/delete?file=#{file}'>" <<
               "<p><input type='submit' value='#{@message["del_file"]}' tabindex='1' accesskey='d' /></p></form>"
    end
    @body << "</body></html>\n"
  end

  #
  # print CGI header (302 moved temporarily)
  #
  def print302(to_next)
    Util.printtrace(to_next) ## debug

    @head << 
        "Content-Type: text/html; charset=UTF-8\r\n" <<
        "Content-Language: en\r\n" <<
        "X-Shingetsu: #{$conf["version"]}\r\n" <<
        "\r\n" <<
    @body << 
        '<?xml version="1.0" encoding="UTF-8"?>' <<
        '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"' <<
        ' "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">' <<
        '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">' <<
        '<head>' <<
        '<meta http-equiv="Content-Language" content="en" />' <<
        '<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />' <<
        '<meta http-equiv="Content-Script-Type" content="text/javascript" />' <<
        "<title>moved</title>" <<
        "<link rev='made' href='mailto:fuktommy\@users.sourceforge.net' />" <<
        "<link rel='contents' href='#{@root}gateway.cgi/list' />" <<
        "</head>" <<
        "<body><p>Click and jump to <a href='#{to_next}'>#{to_next}</a></p>\n" <<
        "<script type='text/javascript'>window.location.href = '#{to_next}';</script>" <<
        "</body></html>\n"
  end

  #
  # format to HTML
  #
  def html_format(arg)
    Util.printtrace(arg) ## debug

    s = arg.dup
    s.gsub!(/<br>/,'<br />')
    s.gsub!(/(&gt;&gt;)([0-9A-Za-z]+)/) {"<a href='#r#{$2}'>#{$&}</a>"}
    s.gsub!(/https?:\/\/[^\x00-\x20"'()<>\[\]\x7F-\xFF]{2,}/) {"<a href=\'#{$&}\'>#{$&}<\/a>"}
    s.gsub!(/\[\[<a.*?>(.*?)\]\]<\/a>/) {"[[#{$1}]]"}
    s.gsub!(/\[\[([^<>]+?)\]\]/) {bracket_link($1)}
    if $conf["redirect"] then
      s.gsub!(%r!<a href='http://shingetsu\.p2p/!, "<a href='/")
    end
    return s
  end

  #
  # read query
  #
  def readQuery()
    Util.printtrace() ## debug

    if ENV["REMOTE_ADDR"] !~ /#{$conf["friendAddr"]}/ then
      print403()
      exit
    elsif ENV["PATH_INFO"] then
      path = ENV["PATH_INFO"].sub(/^\//, "")
    else
      path = ""
    end
  
    if ENV["REQUEST_METHOD"] == "POST" then
      read(STDIN, input, ENV["CONTENT_LENGTH"]) ## ToDo
    else
      input = ENV["QUERY_STRING"]
    end
    return [path, input]
  end
  private :readQuery

  #
  # delete record form
  #
  def deleteRecordForm(file, buf, stamp, id, arg)
    Util.printtrace(file, buf, stamp, id, arg) ## debug

    buf.gsub!(/</, "&lt;")
    buf.gsub!(/>/, "&gt;")
    @body << 
        "<p>#{@message["del_record_q"]}</p>" <<
        "<form method='post' action='#{ENV["SCRIPT_NAME"]}'><p>" <<
        "<input type='hidden' value='xdelete' name='cmd' />" <<
        "<input type='hidden' value='#{file}' name='file' />" <<
        "<input type='hidden' value='#{stamp}' name='stamp' />" <<
        "<input type='hidden' value='#{id}' name='id' />" <<
        "<input type='submit' value='#{@message["yes"]}' name='submit' tabindex='1' accesskey='w' />" <<
        " <input type='checkbox' value='dopost' name='dopost' tabindex='2' accesskey='s' />" <<
        "#{@message["send"]}" <<
        " <input type='checkbox' value='error' name='error'" <<
        " checked='checked' tabindex='2' accesskey='e' />" <<
        "#{@message["error"]}<br />" <<
        " #{@message["name"]}:<input name='name' size='15' value='' tabindex='3' accesskey='n' />" <<
        " #{@message["signature"]}:<input type='password' name='passwd' size='15' value='' tabindex='4' accesskey='p' />" <<
        " #{@message["comment"]}:<input name='message' size='15' value='' tabindex='5' accesskey='m' />" <<
        "<input type='hidden' name='mode' value='#{arg["mode"]}' />" <<
        "</p></form>\n" <<
        "<p>#{buf}</p>"
  end

  #
  # delete file form
  #
  def deleteFileForm(file, bufs)
    Util.printtrace(file, bufs) ## debug

    @body << 
        "<p>#{@message["del_file_q"]} " <<
        "<a href='#{ENV["SCRIPT_NAME"]}?cmd=xdelete&amp;file=#{file}'>" <<
        "#{@message["yes"]}</a></p>\n"
    if file =~ /_/ then
      @body << "<p>" << file_decode(file) << "(#{@message[$`]})</p>\n"
    end
    bufs.each {|item|
      item.chomp!
      item.gsub!(/</, "&lt;")
      item.gsub!(/>/, "&gt;")
      @body << "<p>#{item}</p>\n"
    }
  end

  #
  # delete record dialog
  #
  def deleteRecordDialog(arg)
    Util.printtrace(arg) ## debug

    printHeader(@message["del_record"])
    file = arg["file"]
    @body << "\n"

    buf = nil
    stamp, id, = arg["record"].split(/\//)
    begin
      IO.foreach("#{@datadir}/#{file}.dat") {|line|
        if line =~ /^#{stamp}<>#{id}<>/ then
          buf = line
          break
        end
      }
    rescue Errno::ENOENT, Errno::EACCES
      @body << "<p>#{file}: #{@message["no_file"]}</p></body></html>\n"
    end

    if buf.nil? then
      @body << "<p>#{@message["no_record"]}</p></body></html>\n"
    else
      deleteRecordForm(file, buf, stamp, id, arg)
      @body << "</body></html>\n"
    end
  end

  #
  # delete file dialog
  #
  def deleteFileDialog(arg)
    Util.printtrace(arg) ## debug

    file = arg["file"]
    printHeader(@message["del_file"])
    @body << "\n"

    bufs = []
    count = 0
    IO.foreach("#{@datadir}/#{file}.dat") {|line|
      bufs << line
      break if count > 1
      count += 1
    }
    deleteFileForm(file, bufs)
    @body << "</body></html>\n"
  rescue Errno::ENOENT, Errno::EACCES
    @body << "<p>#{file}: #{@message["no_file"]}</p></body></html>\n"
  end

  #
  # delete file
  #
  def deleteFile(file)
    Util.printtrace(file) ## debug

    File.unlink("#{@datadir}/#{file}.dat")
    stat = CacheStat.list() # Hash
    stat.delete(file)
    CacheStat.sync(stat)
    print302("#{@root}list.cgi")
  rescue
    @head << "Content-Type: text/plain\r\n\r\n"
    @body << "failed. #{$!}\n"
  end

  #
  # delete record
  #
  def deleteRecord(file, stamp, id, arg)
    Util.printtrace(file, stamp, id, arg) ## debug

    if ! Cache.removeRecord(file, stamp, id) then
      @head << "Content-Type: text/plain\r\n\r\n"
      @body << "failed. #{$!}\n"
    elsif arg["dopost"] then  # string
      postDeleteMessage(file, stamp, id, arg)
    end

    if arg["mode"] && (arg["mode"] != "") then
      xfile = str_encode(file_decode(file))
      print302("#{ENV["SCRIPT_NAME"]}/#{xfile}")
    end
  end

  #
  # delete dialog
  #
  def deleteDialog(arg)
    Util.printtrace(arg) ## debug

    if arg["file"].nil? then
      print404()
      exit
    elsif arg["record"] then  ## ????
      deleteRecordDialog(arg)
    else
      deleteFileDialog(arg)
    end
  end

  #
  # check signature
  #
  def checkSign(arg)
    Util.printtrace(arg) ## debug

    unless (arg["file"] && arg["stamp"] && arg["id"]) then
      print404()
      exit
    end

    file = arg["file"]
    touch(file)

    IO.foreach("#{@datadir}/#{file}.dat") {|line|
      hrec = rec(line)
      next unless (hrec["stamp"] == arg["stamp"]) && (hrec["id"] == arg["id"])
      if ! (hrec["pubkey"] && (hrec["pubkey"] != "") &&
            hrec["sign"  ] && (hrec["sign"  ] == "") &&
            hrec["target"] && (hrec["target"] == "")) then
        @head << "Content-Type: text/plain; charset=UTF-8\r\n" <<
                       "X-Shingetsu: #{$conf["version"]}\r\n" <<
                       "\r\n"
        @body << "#{@message["not_signed"]}\n#{line}\n"
      elsif Signature.check(hrec) then
        line.gsub!(/</, "&lt;")
        line.gsub!(/>/, "&gt;")
        trip = Signature.pubkey2trip(hrec["pubkey"])
        type, filename = file.split(/_/, 2)
        printHeader(@message["check_sign"])
        @body << "<p>#{@message["protected"]}(#{hrec["target"]}).</p>" <<
                       "<form method='get' action='#{@root}gateway.cgi/edittrust'><p>" <<
                       "<input type='submit' name='button' value='#{@message["add"]}' tabindex='1' accesskey='a' />" <<
                       "/<input type='submit' name ='button' value='#{@message["delete"]}' tabindex='2' accesskey='d' />" <<
                       "<input type='hidden' name='trip' value='#{trip}' />" <<
                       "<input type='hidden' name='type' value='#{type}' />" <<
                       "<input type='hidden' name='file' value='#{filename}' />" <<
                       "#{@message["to_from_trust_list"]}: #{trip}</p></form>" <<
                       "<p>#{line}</p>" <<
                       "</body></html>"
        exit
      else
        @head << "Content-Type: text/plain; charset=UTF-8\r\n" <<
                       "X-Shingetsu: #{$conf["version"]}\r\n" <<
                       "\r\n"
        @body << "#{@message["wrong_sign"]}.\n#{line}\n"
        exit
      end
    }
    print404(file)
  rescue Errno::ENOENT, Errno::EACCES
    raise "#{@datadir}/#{file}.dat: failed to open. #{$!}#{$@}"
  end

  #
  # new element form
  # opt["file"], opt["checked"], opt["dopost"]
  #
  def newElementForm(opt = nil)
    Util.printtrace(opt) ## debug

    opt = {} if opt.nil?
    opt["checked"] = "" if opt["checked"].nil?
    file = opt["file"] || ""

    @body << 
        "<form method='get' action='#{ENV["SCRIPT_NAME"]}'><p>" <<
        "<input type='hidden' name='cmd' value='new' />" <<
        "<input type='hidden' name='file' value='#{file}' />" <<
        "<input type='submit' value='#{@message["new"]}' name='submit' tabindex='1' accesskey='w' />" <<
        "<input type='radio' value='list' name='type' tabindex='2' accesskey='t' />" <<
        "#{@message["list"]}" <<
        " <input type='radio' value='thread' name='type' tabindex='2' accesskey='t' />" <<
        "#{@message["thread"]}" <<
        " <input type='radio' value='note' name='type' tabindex='2' accesskey='t' />" <<
        "#{@message["note"]}"
    if opt["dopost"] then # string
      @body << " <input type='checkbox' value='dopost' name='dopost'" <<
                     " #{opt["checked"]} tabindex='2' accesskey='s' />" <<
                     "#{@message["send"]}"
    end
    @body << "<br />#{@message["name"]}: <input name='name' size='19' value='' tabindex='3' accesskey='t' />" <<
                   "</p></form>"
  end

end
