#!/usr/bin/ruby -Ku

require "syslog"
require "tserver"
require "util"
require "config"

def log_debug(str)
  Syslog.debug("%s", str) if Syslog.opened? and $conf["debug"]
end

def log_info(str)
  Syslog.info("%s", str) if Syslog.opened?
end

def log_notice(str)
  Syslog.notice("%s", str) if Syslog.opened?
end

def log_warn(str)
  Syslog.warning("%s", str) if Syslog.opened?
end

def log_err(str)
  Syslog.err("%s", str) if Syslog.opened?
end

class THamayu

  class Error < StandardError
  end

  def start()
    ts = nil
    $already_forked = false
    loop = true
    while loop do
      load_config

      TServer.logging = $conf["debug"] ? (Syslog.opened? ? :syslog : true) : nil
      if $conf["port"] != @port then
        ts.close if ts
        ts = TServer.new(*($conf["port"].to_s.split(/:/)))
        @port = $conf["port"]
      end

      ts.handle_signal = false
      ts.on_child_start = Proc.new do
        Process.uid  = $uid if $uid
        Process.euid = $uid if $uid
        Process.gid  = $gid if $gid
      end

      unless @signal_setup then
        trap "HUP" do
          log_notice "THamayu daemon reload"
          ts.stop
        end
        trap "TERM" do
          loop = false
          ts.stop
        end
        trap "INT" do
          ts.interrupt
          loop = false
          ts.stop
        end
        @signal_setup = true
      end

      if $conf["error-log"] then
        File.open($conf["error-log"], "a") do |f|
          STDERR.reopen f
          STDERR.sync = true
        end
      end
      if not $conf["fg"] and not $already_forked then
        if fork then
          exit
        end
        Process.setsid
        $already_forked = true
        log_notice "THamayu daemon start"
      end
      File.unlink @pid_file rescue nil if @pid_file
      @pid_file = nil
      if $conf["pid-file"] and not $conf["pid-file"].empty? then
        File.open($conf["pid-file"], "w") do |f| f.puts $$ end
        @pid_file = $conf["pid-file"]
      end

      ts.min_servers = $conf["min-servers"]
      ts.max_servers = $conf["max-servers"]
      ts.max_use     = $conf["max-use"]
      ts.max_idle    = $conf["max-idle"]

      ts.start do |conn|
        begin
          Conn.new conn
        rescue THamayu::Error
          log_err "#{$!}"
          exit 1
        rescue
          log_err "#{$@[0]}: #{$!}"
          STDERR.puts Time.now.inspect
          STDERR.flush
          raise
        end
      end
    end
    File.unlink @pid_file rescue nil if @pid_file
    log_notice "THamayu daemon stop"
  end

  def exec_client()
    @pid_client = fork {
      # child process
      Signal.trap("TERM", "DEFAULT")
      Signal.trap("HUP",  "DEFAULT")
      Signal.trap("CHLD", "DEFAULT")
      exec "./cgi_client.rb"
    }
  end

  def kill_client()
    if defined?(@pid_client) && (@pid_client > 0) then
      Process.kill("TERM", @pid_client)
      @pid_client = 0
    end
  end

  def load_config()
    if $conf["user"] then
      if $conf["user"] =~ /\A\d+\Z/ then
        $uid = $conf["user"].to_i
      else
        require "etc"
        $uid = Etc.getpwnam($conf["user"]).uid
      end
    end
    if $conf["group"] then
      if $conf["group"] =~ /\A\d+\Z/ then
        $gid = $conf["group"].to_i
      else
        require "etc"
        $gid = Etc.getgrnam($conf["group"]).gid
      end
    end
    if not $conf["syslog"] or $conf["syslog"] == "none" then
      Syslog.close if Syslog.opened?
    else
      fac = eval "Syslog::LOG_#{$conf["syslog"].upcase}"
      if Syslog.opened? then
        Syslog.reopen(File.basename($0), nil, fac)
      else
        Syslog.open(File.basename($0), nil, fac)
      end
    end
  end

  class Conn
    def initialize(sock)
      $conf["debugid"] = "http" # debug

      Util.debug("\033[1;36m==Connect==\033[0m") ## debug

      @sock       = sock
      @user       = nil
      @start_time = Time.now
      @range      = 0
      @gzip       = true
      dummy, @client_port, @client_hostname, @client_ipaddr = @sock.peeraddr

      Signal.trap("ALRM") { 
        raise "#{@client_ipaddr}: server timeout" 
      }

      ENV["REMOTE_HOST" ] = @client_hostname
      ENV["REMOTE_ADDR" ] = @client_ipaddr
      ENV["REMOTE_PORT" ] = @client_port.to_s
      ENV["SCRIPT_NAME" ] = nil
      ENV["QUERY_STRING"] = nil
      ENV["PATH_INFO"   ] = nil
      ENV["HTTP_HOST"   ] = ""

      # read Header
      str_request = readHead()

      if ENV["REQUEST_METHOD"] == "OPTIONS" then
        printHead("200 OK")
        printLine("Allow: GET, HEAD, POST, OPTIONS\r\n",
                  "Content-Length: 0\r\n\r\n")
        return

      elsif ENV["REQUEST_METHOD"] !~ /^(GET|HEAD|POST)$/ then
        printHead("501 Not Implemented")
        printLine("\r\n")
        return
      end

      # request URI
      parse_uri()

      # output logs
      date = Time.now.localtime.to_i
      Util.debug("#{@client_ipaddr}<>#{ENV["SCRIPT_NAME"]}<>#{str_request}<>#{ENV["HTTP_REFERER"]}<>#{ENV["HTTP_USER_AGENT"]}")

      @gzip = false if (ENV["SCRIPT_NAME"]) &&
                       (ENV["SCRIPT_NAME"] != "/server.cgi" ) &&
                       (ENV["SCRIPT_NAME"] != "/hosting.cgi")

      case ENV["SCRIPT_NAME"]
        when "/server.cgi"
          require "cgi_server"
          script = ServerCGI.new
        when "/gateway.cgi"
          require "cgi_gateway"
          script = GatewayCGI.new
        when "/list.cgi"
          require "cgi_list"
          script = ListCGI.new
        when "/note.cgi"
          require "cgi_note"
          script = NoteCGI.new
        when "/thread.cgi"
          require "cgi_thread"
          script = ThreadCGI.new
        else
          scrpit = nil
      end

      #
      # execute script
      #
      if script then
        # path
        if script.class != "ServerCGI" then
          if ENV["REMOTE_ADDR"] !~ /#{$conf["friendAddr"]}/ then
            print403()
            exit
          end
        end

        Util.debug("PATH_INFO    =", ENV["PATH_INFO"]) ## debug
        Util.debug("QUERY_STRING =", ENV["QUERY_STRING"]) ## debug

        if ENV["PATH_INFO"] && (ENV["PATH_INFO"] != "") then
          path = ENV["PATH_INFO"].dup
          path.sub!(/^\//, "")
        elsif ENV["QUERY_STRING"] && (ENV["QUERY_STRING"] != "") then
          path = ENV["QUERY_STRING"].dup
        else
          path = ""
        end

        # read input
        if ENV["REQUEST_METHOD"] == "POST" then
          len = ENV["CONTENT_LENGTH"].to_i
          if len > 0 then
            input = readBody(len)
            raise "404" if input.nil?
          end
        else
          input = ENV["QUERY_STRING"]
        end

        # execute cgi
        ENV["SCRIPT_FILENAME"] = $conf["cgiDir"] + ENV["SCRIPT_NAME"]
        begin
          script.execute(path, input)
        rescue Gateway::Error
        end

        $conf["debugid"] = "http" # debug
        Util.debug("body = ", script.body) # debug

        # result
        printResult(script.head, script.body)
      elsif ENV["REQUEST_URI"] =~ /^\/[0-9A-Za-z._]+$/ then
        ufile = $&
        printFile(ufile)
      else
        raise "404"
      end
    rescue
      printHead("404 Not Found")
      printLine("\r\n",
                "Bad arguments or No data.\n",
                "Try later.\n")
      raise
    ensure
      Util.debug("\033[1;36m==Disconnect==\033[0m") ## debug
      sleep(2)
#      @sock.close rescue nil
    end

    def readHead()
      Util.printtrace() ## debug

      ENV["REQUEST_METHOD"] = ""
      ENV["CONTENT_LENGTH"] = nil

      request_str = ''
      alarm_on()
      @sock.each_line {|line|
        line.gsub!(/[\n\r]/,'')
        break if line == ""

        line =~ /^(\S+)\s+/
        key = $1.upcase
        val = $'
        next if key.nil? || val.nil?

        key.gsub!(/:$/, '')

        if ENV["REQUEST_METHOD"] == "" then
          request_str = line
          val, = val.split(/\s/, 2)
          ENV["REQUEST_METHOD"] = key
          ENV["REQUEST_URI"   ] = val.gsub(%r!^//+!, "/")
          next
        end

        case key
          when "ACCEPT-ENCODING"
            if (val =~ /.*gzip/i) &&
               (ENV["REMOTE_ADDR"] !~ /#{$conf["adminAddr"]}/) then
              @gzip = true if val =~ /.*gzip/i
            end

          when "RANGE"
            if val =~ /^bytes\s*=\s*(\d+)\s*-/ then
              @range = $1.to_i 
            end

          when "CONTENT-LENGTH"
            if val =~ /(\d+)/ then
              ENV["CONTENT_LENGTH"] = $1
            end

          when "CONTENT-TYPE"
            if val =~ /([^;]+)/ then
              ENV["CONTENT_TYPE"] = $1
            end

          else
            ENV["HTTP_" + key.gsub(/-/, "_")] = val
        end
      } # --> @sock.each_line {|line|
      alarm_off()
    ensure
      return request_str
    end


    def parse_uri()
      Util.printtrace() ## debug

      if ENV["REQUEST_URI"] =~ %r!^[a-z]+://([^/]+)! then
        ENV["HTTP_HOST"  ] = $1
        ENV["REQUEST_URI"] = $'
      end

      if ENV["REQUEST_URI"] == "" then
        ENV["REQUEST_URI"] = "/"
      end

      if ENV["REQUEST_URI"] =~ /^\/([0-9a-zA-Z]+)\.cgi/i then
        ENV["SCRIPT_NAME"] = "/#{$1}.cgi"
        ENV["PATH_INFO"  ] = $'

        if (ENV["PATH_INFO"] =~ /\?/) then
          ENV["QUERY_STRING"] = $'
          ENV["PATH_INFO"   ] = $`
        else
          ENV["QUERY_STRING"] = ""
        end
        require "uri"
        ENV["PATH_INFO"] = URI.decode(ENV["PATH_INFO"])

      elsif ENV["REQUEST_URI"] == "/" then
        ENV["SCRIPT_NAME" ] = $conf["index"]
        ENV["QUERY_STRING"] = ""
      end
    end

    def readBody(len)
      Util.printtrace(len) ## debug

      body = ''

      if len > ($conf["filelimit"] * 1024 * 1024) then
        return nil
      else
        bsize = $conf["bufsize"]
        buf = " " * bsize   # memory alloc
        while len > bsize do
          alarm_on()
          @sock.read(bsize, buf)
          body << buf
          len  -= bsize
          alarm_off()
        end
        if len > 0 then
          alarm_on()
          @sock.read(len, buf)
          body << buf
          alarm_off()
        end
      end
      return body
    ensure
      alarm_off()
    end

    def alarm_on()
      Util.alarm($conf["timeout"])
    end

    def alarm_off()
      Util.alarm(0)
    end

    def printHead(status)
      Util.printtrace() ## debug

      alarm_on()
      @sock.print "HTTP/1.0 #{status}\r\n",
                  "Server: #{$conf["version"]}\r\n",
                  "Connection: Close\r\n"
      alarm_off()
    end

    def printLine(*arg)
      alarm_on()
      @sock.print arg
      alarm_off()
    end

    def printResult(head, body)
      Util.printtrace() ## debug

      # Header
      len = body.size

      if @range > 0 then
        all  = body.size
        body = body[@range..-1]  # -1 = last
        len  = body.size
        printHead("206 Partial content")
        printLine("Content-Range: bytes #{@range}-#{(all-1).to_s}/#{all}\r\n",
                  "Content-Length: #{len}\r\n")

      elsif @gzip then
        require "extern"
        buf_gzip = Extern.xgzip(body)
        if buf_gzip.size < body.size then
          body = buf_gzip
          len = body.size
          printHead("200 OK")
          printLine("Content-Encoding: gzip\r\n",
                    "Content-Length: #{len}\r\n")
        else
          printHead("200 OK")
          printLine("Content-Length: #{len}\r\n")
        end

      else
        printHead("200 OK")
        printLine("Content-Length: #{len}\r\n")
      end

      if head != "" then
        if len < (100 * 1024) then
          printLine("Cache-Control: no-cache, no-store, must-revalidate\r\n")
          printLine("Expires: -1\r\n")
          printLine("Last-Modified: " + Time.new.gmtime.strftime("%a, %d %b %Y %H:%M:%S GMT") + "\r\n")
        end
        printLine(head)
        printLine("\r\n") if head[-4..-1] != "\r\n\r\n"
      end

      if ENV["REQUEST_METHOD"] != "HEAD" then
        bsize = $conf["bufsize"]
        idx = 0
        while len > bsize do
          alarm_on()
          printLine(body[idx, bsize])
          alarm_off()
          idx += bsize
          len -= bsize
        end
        alarm_on()
        printLine(body[idx, len]) if len > 0
        alarm_off()
      end
    end

    def printFile(file)
      Util.printtrace(file) ## debug

      file.sub!(/^\/+/, '')   # "/" --> "+"  (replace once)
      file   =~ /\.([^.]+)$/  # ExtractFileExt
      suffix = $1
      suffix = "txt" if suffix.nil?

      type    = $conf["mimeType"][suffix].dup
      filedir = $conf["filedir"].dup
      type    = "text/plain" if type.nil?

      fsize = 0
      begin
        fsize = FileTest.size("#{filedir}/#{file}")   
        # not file exist --> raise exception
      rescue
        printHead("404 Not Found")
        printLine("\r\nBad arguments or No data.\n",
                   "Try later.\n")
        return
      end

      File.open("#{filedir}/#{file}", "r") {|f|
        printHead("200 OK")
        printLine("Content-Type: #{type}\r\n")
        if ENV["REQUEST_METHOD"] == "HEAD" then
          printLine("Content-Length: 0\r\n", "\r\n")
          return
        elsif @range > 0 then
          len = fsize - @range
          f.seek(@range)
          printLine("Content-Range: bytes #{@range}-#{(fsize-1).to_s}/#{all}\r\n",
                    "Content-Length: #{len}\r\n",
                    "\r\n")
        else
          printLine("Content-Length: #{fsize.to_s}\r\n",
                    "\r\n")
        end
        buf = "x" * $conf["bufsize"]
        while f.read($conf["bufsize"], buf)
          printLine(buf)
        end
      }
    end
  end
end

#
# Main
#
begin
  initNode = []
  while !ARGV.empty? do
    s = ARGV.shift
    if s == "-v" then
      $conf["debug"] = true
    elsif s == "-V" then
      $conf["debug"] = false
    elsif s == "-p" then
      $conf["port"] = ARGV.shift
    elsif s == "-f" then
      $conf["fg"] = true
    elsif s == "-C" then
      $conf["noclient"] = true
    elsif s == "-i" then
      initNode << ARGV.shift
    end
  end

  ENV["SERVER_SOFTWARE"  ] = "#{$conf["version"]}"
  ENV["GATEWAY_INTERFACE"] = "CGI/1.1"
  ENV["SERVER_PROTOCOL"  ] = "HTTP/1.1"
  ENV["SERVER_PORT"      ] = $conf["port"].to_s
  ENV["HTTP_PORT"        ] = $conf["port"].to_s
  ENV["SERVER_NAME"      ] = ""
  ENV["HTTP_HOST"        ] = ""
  ENV["X_INIT_NODE"      ] = initNode.join("+") unless initNode.empty?

  hm = THamayu.new
  hm.exec_client()
  hm.start()
ensure
  hm.kill_client() if hm
end
